from .highlander import highlander
from .exceptions import RedisKeyError
from .util import Serializer, Deserializer, serializers, deserializers, serializer, deserializer
from .property import Property
from .collection import Collection, Set, List, SortedSet, Hash
from .model import Model, ModelType
from .database import Database, DatabaseType



import pickle
from trol import RedisKeyError, Serializer, Deserializer


class Property:
    """A property object which emulates a remote object as local

    The expectation is that this Property object is embedded as a class level attribute
    The class it is embedded in should define `redis` and `key` attrbutes to define where this property is stored

    Note:
        This class is intended to be used with trol.Model subclasses, but can work on it's own

    Args:
        autocommit (bool): Commit the local value to Redis on every assignment
            If set to None, assignments will check the holder object for an autocommit value, and default to True if not present
            Useful if you don't want to worry about forgetting to commit
        alwaysfetch (bool): Fetch the latest value from the database
            If set to None, gets will check the holder object for an alwaysfetch value, and default to True if not present
            Useful if you have multiple threads or processes frequently modifying the database
        serializer (Callable[[object], bytes]): A function or callable which will be used for serializing the value before storage in Redis
            Although bytes is the most general output type, this function may also output `str`, `int`, or any other type redis=py will accept
            Default is `pickle.dumps`, which is not human readable, but will work with most python object
        deserializer (Callable[[bytes], object]): A function or callable which will be used for deserializing the value after reciving it from redis
            Default is `pickle.loads`, the counterpart for the deafault serializer

    Args:
        autocommit (bool): Sets the autocommit attribute. Defaults to None
        alwayfetch (bool): Sets the alwaysfetch attribute. Defaults to None
        serializer (Callable[[object], bytes]): Sets the serializer attribute. Defaults to None
        deserializer (Callable[[bytes], object]): Sets the deserializer attribute. Defaults to None
    """
    class Null:
        """A class to act as an indicator value"""

    null = Null()
    """Property.Null: An indicator field to show the value needs to be fetched"""

    @staticmethod
    def mangle(name):
        """Creates a mangled version of the inputted name

        Mangling is the process of changing an attribute name in such a way that it will liekly not collide with other attributes

        Args:
            name (str): The name which should be mangled

        Returns:
            str: The mangled name
        """
        return "_rtol_property_{}".format(name)

    def __init__(self, name=None, typ=None, autocommit=None, alwaysfetch=None, serializer=None, deserializer=None):
        self.name = name
        self.autocommit = autocommit
        self.alwaysfetch = alwaysfetch

        self._typ = typ
        if serializer is None:
            if typ is None:
                self.serialize = pickle.dumps
            else:
                self.serialize = Serializer(typ)
        else:
            self.serialize = serializer

        if deserializer is None:
            if typ is None:
                self.deserialize = pickle.loads
            else:
                self.deserialize = Deserializer(typ)
        else:
            self.deserialize = deserializer

    def __get__(self, obj, typ=None):
        if obj is None:
            return self

        value = self.value(obj)

        if value is self.null or self.alwaysfetch or (self.alwaysfetch is None and getattr(obj, 'alwaysfetch', False)):
            value = self.fetch(obj)

        if value is self.null:
            raise RedisKeyError(self.key(obj))

        return value

    def __set__(self, obj, value):
        self.set(obj, value)

        if self.autocommit or (self.autocommit is None and getattr(obj, 'autocommit', True)):
            self.commit(obj)

    @property
    def name(self):
        """``str``: The name for this property, which will be used to determine the key when bound to a Model"""
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    def fetch(self, obj):
        """Retrieves and sets the value of this property

        Args:
            obj (object): This property's holder

        Returns:
            bytes: The data retrieved or None in the case of a key not found
        """
        response = obj.redis.get(self.key(obj))

        if response is not None:
            value = self.deserialize(response)
        else:
            value = self.null

        self.set(obj, value)
        return value

    def commit(self, obj):
        """Commits this properties value to Redis

        Does nothing if the value is Property.null, which means there is nothing to write

        Args:
            obj (object): This property's holder

        Returns:
            bool: True if the set transaction was successful. False otherwise
        """
        value = self.value(obj)
        if self.value(obj) is self.null:
            return True

        return obj.redis.set(self.key(obj), self.serialize(value))

    def delete(self, obj):
        """Deletes the key of this property from Redis

        Args:
            obj (object): This property's holder

        Returns:
            bool: True if the delete transaction was successful. False otherwise
        """
        ok = obj.redis.delete(self.key(obj))
        if ok:
            self.set(obj, None)
        return ok

    def invalidate(self, obj):
        """Invalidates the local value to indicate a fetch must be done

        Args:
            obj (object): This property's holder
        """
        self.set(obj, self.null)

    def key(self, obj):
        """Gets the key where this property's data exists  in Redis

        The key for this property is the key of it's holder, with `:<name>` appended

        Args:
            obj (object): This propery's holder

        Returns:
            str: The key which can be used to get this property's data from Redis
        """
        if obj.key is not None:
            return "{}:{}".format(obj.key, self.name)
        else:
            return self.name

    def value(self, obj):
        """Gets the value stored in the holder obj

        Sets the property value attribute in the holder if it does not already exist

        Args:
            obj (object): This propery's holder

        Returns:
            object: The local value of this property
        """
        try:
            value = getattr(obj, self.mangle(self.name))
        except AttributeError:
            self.set(obj, self.null)
            value = self.null
        return value

    def set(self, obj, value):
        """Sets the value stored in the holder obj

        Args:
            obj (object): This propery's holder
            value (object): The value to set
        """
        setattr(obj, self.mangle(self.name), value)



import unittest
from trol import Model, Property, Set, Database

class TestDatabase(unittest.TestCase):
    def test_init(self):
        canary = object()

        class X(Database):
            redis = canary

            prop = Property()

            collect = Set()

            class Mod(Model):
                pass

        self.assertIs(X.prop, X._rtol_properties['prop'])
        self.assertEquals(X.prop.name, 'prop')

        self.assertIs(X.collect, X._rtol_collections['collect'])
        self.assertEquals(X.collect.name, 'collect')

        self.assertIs(X.Mod, X._rtol_models['Mod'])
        mod = X.Mod()
        self.assertIs(mod.redis, canary)



import docker
import unittest
import pickle
from redis import StrictRedis
from trol import Property, RedisKeyError
from .common import ensure_redis_is_online


class A:
    def __init__(self, redis=None, autocommit=False):
        self.redis = redis

    alpha = Property("abc", autocommit=False)


class OfflinePropertyTests(unittest.TestCase):
    def test_mangle(self):
        nameA, nameB = "fred", "wilma"
        self.assertNotEqual(Property.mangle(nameA), nameA)
        self.assertNotEqual(Property.mangle(nameA), Property.mangle(nameB))

    def test_key(self):
        class X:
            prop = Property("prop")
            key = None

        x = X()
        self.assertEquals(X.prop.key(x), "prop")

        x.key = "xkey"
        self.assertEquals(X.prop.key(x), "xkey:prop")

    def test_value_and_set(self):
        class X:
            prop = Property("prop")
            key = None

        x = X()
        self.assertIs(X.prop.value(x), X.prop.null)
        canary = object()
        X.prop.set(x, canary)
        self.assertIs(X.prop.value(x), canary)


class OnlinePropertyTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.container_token = ensure_redis_is_online()
        cls.redis = StrictRedis(host='localhost', port=6379, db=0)

    def setUp(self):
        self.redis.flushall()

    def test_self(self):
        """Make sure the docker container can be booted and redis can be connected to

        If this test fails, then the rest surely won't pass
        """
        self.redis.set("key", "value")
        self.assertEquals(self.redis.get("key").decode('utf-8'), "value")
        self.redis.delete("key")

    def test_commit_nothing(self):
        class X:
            prop = Property("p")
            key = "xkey"

        x = X()
        key = X.prop.key(x)
        self.assertIsNone(self.redis.get(key))

        self.redis.set(key, "canary")
        self.assertTrue(X.prop.commit(x))
        self.assertEquals(self.redis.get(key).decode('utf-8'), "canary")

    def test_commit_something(self):
        class X:
            redis = self.redis
            prop = Property("p")
            key = "xkey"

        x = X()
        key = X.prop.key(x)
        X.prop.set(x, "something")
        self.assertIsNone(self.redis.get(key))

        self.redis.set(key, "canary")
        self.assertTrue(X.prop.commit(x))
        self.assertEquals(pickle.loads(self.redis.get(key)), "something")

    def test_setter_autocommit(self):
        class X:
            redis = self.redis
            prop = Property("p", autocommit=True)
            key = "xkey"

        x = X()
        key = X.prop.key(x)
        self.assertIsNone(self.redis.get(key))
        self.redis.set(key, "canary")

        x.prop = "something"
        self.assertEquals(pickle.loads(self.redis.get(key)), "something")

    def test_setter_no_autocommit(self):
        class X:
            redis = self.redis
            prop = Property("p", autocommit=False)
            key = "xkey"

        x = X()
        key = X.prop.key(x)
        self.assertIsNone(self.redis.get(key))
        self.redis.set(key, "canary")

        x.prop = "something"
        self.assertEquals(self.redis.get(key).decode('utf-8'), "canary")
        self.assertTrue(X.prop.commit(x))
        self.assertEquals(pickle.loads(self.redis.get(key)), "something")

    def test_redis_key_error(self):
        class X:
            redis = self.redis
            prop = Property("p", alwaysfetch=True)
            key = "xkey"

        x = X()
        key = X.prop.key(x)

        with self.assertRaises(RedisKeyError) as ex:
            x.prop

        self.assertEquals(ex.exception.key, key)
        self.assertIn(key, str(ex.exception))

    def test_getter_alwaysfetch(self):
        class X:
            redis = self.redis
            prop = Property("p", alwaysfetch=True)
            key = "xkey"

        x = X()
        key = X.prop.key(x)
        with self.assertRaises(RedisKeyError):
            x.prop
        self.assertIsNone(self.redis.get(key))

        self.redis.set(key, pickle.dumps("canary"))
        self.assertEquals(x.prop, "canary")

    def test_getter_no_alwaysfetch(self):
        class X:
            redis = self.redis
            prop = Property("p", alwaysfetch=False)
            key = "xkey"

        x = X()
        key = X.prop.key(x)

        with self.assertRaises(RedisKeyError):
            x.prop
        self.assertIsNone(self.redis.get(key))

        self.redis.set(key, pickle.dumps("canary"))
        self.assertEquals(x.prop, "canary")

        self.redis.set(key, pickle.dumps("sparrow"))
        self.assertEquals(x.prop, "canary")

        X.prop.invalidate(x)
        self.assertEquals(x.prop, "sparrow")

    def test_delete(self):
        class X:
            redis = self.redis
            prop = Property("p")
            key = "xkey"

        x = X()
        key = X.prop.key(x)

        self.redis.set(key, "canary")
        self.assertTrue(X.prop.delete(x))
        self.assertIsNone(self.redis.get(key))
        self.assertFalse(X.prop.delete(x))



from functools import wraps
from trol import Property, Collection, deserializer, serializer
import weakref

"""Provides the Model and ModelType classes, which are the basic blocks of trol

The Model class is what trol data models derive from.
The examples in these docs will use the following data model

"""

_all_models = weakref.WeakValueDictionary()


class ModelType(type):
    """A metaclass which provides awareness of properties and collections

    This type embeds a dict of :obj:`Property` which tracks any properties assigned at class load.
    It assigns the names of properties who are unamed
    This type embeds a dict of :obj:`Collection` which tracks any collections assigned at class load
    It will assign the names to any collections which did not have their names assigned
    """
    def __init__(cls, *args, **kwargs):
        cls._rtol_properties = dict()
        cls._rtol_collections = dict()

        for attrname, attr in cls.__dict__.items():
            if isinstance(attr, Property):
                cls._rtol_properties[attrname] = attr

                if attr._name is None:
                    attr._name = attrname

            if isinstance(attr, Collection):
                cls._rtol_collections[attrname] = attr

                if attr._name is None:
                    attr._name = attrname

        _all_models[cls.__name__] = cls

        super().__init__(*args, **kwargs)


class Model(metaclass=ModelType):
    """A class to support object oriented Redis communication

    Attributes:
        autocommit (bool): If `True`, properties will be immediatly commited to Redis when modified. Default is `True`
            This attribute can be overriden for a single property by setting that properties `autocommit` attribute
        alwaysfetch (bool): If `True`, property values will be fetched from Redis on every access. Deafault is `False`
            This attribute can be overriden for a single property by setting that properties `alwaysfetch` attribute
    """
    _redis = None
    _key = None
    _model_name = None

    autocommit = True
    alwaysfetch = False

    @property
    def key(self):
        """str: Redis key which prefixes any properties stored under this Model instance

        By default this is {model_name}:{id}

        Changing the key of an instance may cause data to be lost.
        It's best to think of these models as a pointer and changing the keys is changing the value of the pointer

        Example:
            TODO: Write a new example

        """
        if self._key is not None:
            return self._key

        return ':'.join((self.model_name, self.id))

    @key.setter
    def key(self, key):
        self._key = key

    @property
    def model_name(self):
        """str: A name for this model which, in addition to its id which identify it in Redis

        By default this is the class name

        Example:
            TODO: Write a new example

        """
        if self._model_name is not None:
            return self._model_name

        return self.__class__.__name__

    @model_name.setter
    def model_name(self, name):
        self._model_name = name

    @property
    def redis(self):
        """Redis: The active Redis connection for this model

        This model can have it's own Redis connection of use connection of the :obj:`Database` that holds it
        """
        if self._redis is not None:
            return self._redis

        # Use try/except rather than if/else for this block because it should succeed if models are correct
        try:
            return self._rtol_database.redis
        except AttributeError:
            return None

    @redis.setter
    def redis(self, redis):
        self._redis = redis

    def invalidate(self, *propnames):
        """Mark properties in this model as invalid and requiring a fetch

        Args:
            *propnames (list[str]): The attribute nanes of properties which should be invalidated.
                If none are provided, the default is to invalidate all propertoes in the model.
        """
        if propnames:
            props = list()
            for propname in propnames:
                props.append(self._rtol_properties[propname])
        else:
            props = self._rtol_properties.values()

        for prop in props:
            prop.invalidate(self)

    def commit(self, *propnames):
        """Saves properties in this model to Redis

        Args:
            *propnames (list[str]): The attribute nanes of properties which should be committed.
                If none are provided, the default is to commits all propertoes in the model.
        """
        if propnames:
            props = list()
            for propname in propnames:
                props.append(self._rtol_properties[propname])
        else:
            props = self._rtol_properties.values()

        mappings = dict()
        for prop in props:
            value = prop.value(self)
            if value is not prop.null:
                mappings[prop.key(self)] = prop.serialize(value)

        self.redis.mset(mappings)

    def delete(self, *propnames):
        """Deletes properties in this model from Redis

        Args:
            *propnames (list[str]): The attribute nanes of properties which should be deleted.
                If none are provided, the default is to delted all propertoes in the model.
        """
        if propnames:
            props = list()
            for propname in propnames:
                props.append(self._rtol_properties[propname])
        else:
            props = self._rtol_properties.values()

        keys = []
        for prop in props:
            keys.append(prop.key(self))

        self.redis.delete(*keys)

        for prop in props:
            prop.set(self, prop.null)

    def update(self, **kwargs):
        """Updates the local values of multiple properties and commits them if autocommit is set

        Args:
            **kwargs (dict[str, object]): Key value pairs where the key is the property name and value is what it should be set to
        """
        commits = list()
        for propname, value in kwargs.items():
            prop = self._rtol_properties[propname]
            prop.set(self, value)

            if prop.autocommit or (prop.autocommit is None and self.autocommit):
                commits.append(propname)

        if commits:
            self.commit(*commits)


_seperator = b'\xfe'
_indicator = b'\xfc'

@serializer(Model)
def serialize_model(model):
    """Serialize a model instance into a key reference

    The model class, id, model_name, and key will be preserved on serialization
    Any custom attributes of the instance will not

    Args:
        model (Model): The model to serialize

    Returns:
        bytes: The key reference
    """
    model_id = getattr(model, 'id', None)
    class_name = model.__class__.__name__.encode('utf-8')
    model_id = _indicator if model_id is None else str(
        model_id).encode('utf-8')
    model_name = _indicator if model._model_name is None else model._model_name.encode(
        'utf-8')
    model_key = _indicator if model._key is None else model._key.encode(
        'utf-8')

    key = (class_name, model_id, model_name, model_key)
    return _seperator.join(key)


class ModelDeserializationError(Exception):
    def __init___(self, key):
        self.key = key

    def __str__(self):
        return "Failed to deserialize '{}' to a Model".format(self.key)

@deserializer(Model)
def deserialize_model(byts):
    """Deserialize a key reference into a model instance

    The model class, id, model_name, and key will be set on deserialization
    Any custom attributes of the instance will not, and __init__ will not be called

    Args:
        bytes: The key reference

    Returns:
        model (Model): The deserialized model
    """
    try:
        pieces = byts.split(_seperator)

        key = []
        for piece in pieces:
            if piece == _indicator:
                key.append(None)
            else:
                key.append(piece.decode('utf-8'))

        cls = _all_models[key[0]]
        inst = cls.__new__(cls)

        if key[1] is not None:
            inst.id = key[1]
        inst._model_name = key[2]
        inst._key = key[3]
        return inst

    except Exception as err:
        raise ModelDeserializationError(byts) from err



# -*- coding: utf-8 -*-
# doctest: +ELLIPSIS
"""Collections which proxy access to Redis storage

This file is Copyright (c) 2010 Tim Medina
Licenced under the MIT License availible in the root of the RTOL source repo
It has been modifed to hit the overal model of RTOL, but is largely the same externally
Thanks to the guys at Redisco for the great work! https://github.com/kiddouk/redisco

Here are the modifications from the origonal Redisco conatainers:
    * Values are serialized and deserialized allowing more arbitrary objects to be stored
    * Comparison operations do not trigger network transfer of container members
    * Set copy does not tigger network transfer of set members
    * Functions which require multiple explicitly use the pipeline to avoid multiple network calls
    * Some functions now use a "scratch-pad" in redis to avoid transfering full collections
    * The ``db`` attribute has been renamed to ``redis`` to match other places in trol
    * There is no default expire time
    * A redis connection must be specified on construction, unless accessed through a Model
    * There is no support for TypedList or NonPersitentList

Unlike properties, collections call out to Redis on every transaction by default
Caching for collections is a trickier prospect, and therefore is not attempted

>>> import trol
>>> from redis import Redis
>>> redis = Redis()

"""

import threading
import collections
import pickle
from trol import Serializer, Deserializer


# Use a guid to make sure that no wil define a colliding key by accident
# Every thread gets their own scratch key to make sure there are not threading errors
_scratch_guid = "c05f145e-a1a8-4742-918c-e01d6d40b02a"


def _scratch_key():
    return "scratch:{}:{}".format(_scratch_guid, threading.get_ident())


def _parse_values(values):
    (_values,) = values if len(values) == 1 else (None,)
    if _values and type(_values) == type([]):
        return _values
    return values


class Collection(object):
    """
    Base class for all collections. This class should not used directctly
    This class provides the ``redis`` attribute
    :members:
    """

    def __init__(self, name=None, redis=None, typ=None, key=None, serializer=None, deserializer=None):
        self._name = name
        self._redis = redis
        self._key = key
        self._threadlocal = None

        self._typ = typ
        if serializer is None:
            if typ is None:
                self.serialize = pickle.dumps
            else:
                self.serialize = Serializer(typ)
        else:
            self.serialize = serializer

        if deserializer is None:
            if typ is None:
                self.deserialize = pickle.loads
            else:
                self.deserialize = Deserializer(typ)
        else:
            self.deserialize = deserializer

    def __get__(self, obj, typ=None):
        if obj is None:
            return self

        if self._key is not None:
            key = self._key
        else:
            if self._name is not None:
                key = ':'.join((obj.key, self._name))
            else:
                raise AttributeError(
                    "{self.__class__.__name__} does not have it's 'name' or 'key' attributes set. If bound to a class which is not a Model, at least one must be set explicitly")

        if self._redis is not None:
            redis = self._redis
        else:
            redis = obj.redis

        return self.__class__(name=self._name, key=key, redis=redis, typ=self._typ, serializer=self.serialize, deserializer=self.deserialize)

    def clear(self):
        """
        Remove the collection from the redis storage

        >>> s = trol.Set('test', redis)
        >>> s.add('1')
        1
        >>> s.clear()
        >>> s.members
        set()


        """
        self.redis.delete(self.key)

    def set_expire(self, time):
        """
        Allow the key to expire after ``time`` seconds.

        >>> s = trol.Set('test', redis)
        >>> s.add("1")
        1
        >>> s.set_expire(1)
        >>> from time import sleep
        >>> sleep(1.5)
        >>> s.members
        set()

        :param time: time expressed in seconds.
        :rtype: None
        """
        self.redis.expire(self.key, time)

    @property
    def name(self):
        """``str``: The name for this collection, which will be used to determine the key if bound to a Model

        >>> class Alpha(trol.Model):
        ...     def __init__(self, ident):
        ...         self.id = ident
        ...
        ...     storage = trol.Set(name="store")
        ...
        >>> a = Alpha('xyz')
        >>> a.key
        'Alpha:xyz'
        >>> a.storage.key
        'Alpha:xyz:store'

        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def redis(self):
        """``redis.Redis``: A connection object for interacting with Redis"""
        if self._redis is not None:
            return self._redis
        else:
            raise AttributeEror(
                "'{self.__class__.__name__}' has no connection set. If not bound to a class, a collection must have it's connection specified".format(self=self))

    @redis.setter
    def redis(self, value):
        self._redis = value

    @property
    def key(self):
        """``str``: The key in Redis which contains the data this object references

        Setting this attribute directly to non-None will override ``name`` in determining ``key``

        >>> s = trol.Set(name='foo')
        >>> s.key
        'foo'
        >>> s.key = 'bar'
        >>> s.key
        'bar'

        """
        if self._key is not None:
            return self._key
        if self.name is not None:
            return self.name
        else:
            raise AttributeEror(
                "'{self.__class__.__name__}' has no name or key set. If not bound to a class, a collection must have either it's key or name attribute specified".format(self=self))

    @key.setter
    def key(self, value):
        self._key = value

    @property
    def pipeline(self):
        """``redis.Pipeline``: A pipeline object to execute buffered commands

        The collection waits until you call this function for the first time to intialize the pipeline
        A unique pipeline will be created for each thread that accesses this property, so it is thread safe
        """
        if self._threadlocal is None:
            self._threadlocal = threading.local()

        try:
            pipe = self._threadlocal.pipeline
        except AttributeError:
            pipe = self.redis.pipeline()
            self._threadlocal.pipeline = pipe
        return pipe


class Set(Collection):
    """
    .. default-domain:: set

    This class represents a Set in redis.
    """

    def __repr__(self):
        """Gets the string representation of this Set

        >>> s = trol.Set('test', redis)
        >>> repr(s)
        "<Set 'test'>"

        """
        return "<%s '%s'>" % (self.__class__.__name__, self.key)

    def sadd(self, *values):
        """
        Add the specified members to the Set.

        :param values: a list of values or a simple value.
        :rtype: integer representing the number of value added to the set.

        >>> s = trol.Set('test', redis)
        >>> s.add([1, 2, 3])
        3
        >>> s.add([4])
        1
        >>> s.members == {1, 2, 3, 4}
        True
        >>> s.clear()

        """
        values = [self.serialize(v) for v in _parse_values(values)]
        return self.redis.sadd(self.key, *values)

    def srem(self, *values):
        """
        Remove the values from the Set if they are present.

        :param values: a list of values or a simple value.
        :rtype: boolean indicating if the values have been removed.

        >>> s = trol.Set('test', redis)
        >>> s.add([1, 2, 3])
        3
        >>> s.srem([1, 3])
        2
        >>> s.clear()

        """
        values = [self.serialize(v) for v in _parse_values(values)]
        return self.redis.srem(self.key, *values)

    def spop(self):
        """
        Remove and return (pop) a random element from the Set.

        :rtype: String representing the value poped.

        >>> s = trol.Set('test', redis)
        >>> s.add("a")
        1
        >>> s.spop()
        'a'
        >>> s.members
        set()

        """
        return self.deserialize(self.redis.spop(self.key))

    def isdisjoint(self, other):
        """
        Return True if the set has no elements in common with other.

        :param other: another ``Set``
        :rtype: boolean

        >>> s1 = trol.Set('key1', redis)
        >>> s2 = trol.Set('key2', redis)
        >>> s1.add(['a', 'b', 'c'])
        3
        >>> s2.add(['c', 'd', 'e'])
        3
        >>> s1.isdisjoint(s2)
        False
        >>> s1.clear()
        >>> s2.clear()

        """
        self.pipeline.sinterstore(_scratch_key(), [self.key, other.key])
        self.pipeline.delete(_scratch_key())
        return self.pipeline.execute()[0] == 0

    def issubset(self, other_set):
        """
        Test whether every element in the set is in other.

        :param other_set: another ``Set`` to compare to.

        >>> s1 = trol.Set('key1', redis)
        >>> s2 = trol.Set('key2', redis)
        >>> s1.add(['a', 'b', 'c'])
        3
        >>> s2.add('b')
        1
        >>> s2.issubset(s1)
        True
        >>> s1.clear()
        >>> s2.clear()

        """
        return self <= other_set

    def __le__(self, other_set):
        self.pipeline.sdiffstore(_scratch_key(), [self.key, other_set.key])
        self.pipeline.delete(_scratch_key())
        return self.pipeline.execute()[0] == 0

    def __lt__(self, other_set):
        """Test whether the set is a true subset of other.

        >>> s1 = trol.Set('key1', redis)
        >>> s2 = trol.Set('key2', redis)
        >>> s1.add(['a', 'b'])
        2
        >>> s2.add(['a', 'b'])
        2
        >>> s2 < s1
        False
        >>> s1.add('c')
        1
        >>> s2 < s1
        True
        >>> s1.clear()
        >>> s2.clear()

        """
        self.pipeline.sdiffstore(_scratch_key(), [self.key, other_set.key])
        self.pipeline.scard(self.key)
        self.pipeline.scard(other_set.key)
        self.pipeline.delete(_scratch_key())
        result = self.pipeline.execute()
        return result[0] == 0 and result[1] != result[2]

    def __eq__(self, other_set):
        """
        Test equality of keys first, then members is they are not equal

        >>> s1 = trol.Set('key1', redis)
        >>> s2 = trol.Set('key2', redis)
        >>> s1.add(['a', 'b'])
        2
        >>> s2.add(['a', 'b'])
        2
        >>> s2 == s1
        True
        >>> s1.add('c')
        1
        >>> s2 < s1
        True
        >>> s1 == s1
        True
        >>> s1.clear()
        >>> s2.clear()

        """
        if other_set.key == self.key:
            return True

        self.pipeline.sunionstore(_scratch_key(), [self.key, other_set.key])
        self.pipeline.scard(self.key)
        self.pipeline.scard(other_set.key)
        self.pipeline.delete(_scratch_key())
        result = self.pipeline.execute()
        return result[0] == result[1] and result[1] == result[2]

    def __ne__(self, other_set):
        return not self.__eq__(other_set)

    def issuperset(self, other_set):
        """
        Test whether every element in other is in the set.

        :param other_set: another ``Set`` to compare to.

        >>> s1 = trol.Set('key1', redis)
        >>> s2 = trol.Set('key2', redis)
        >>> s1.add(['a', 'b', 'c'])
        3
        >>> s2.add('b')
        1
        >>> s1.issuperset(s2)
        True
        >>> s1.clear()
        >>> s2.clear()

        """
        return self >= other_set

    def __ge__(self, other_set):
        """Test whether every element in other is in the set."""
        return other_set <= self

    def __gt__(self, other_set):
        """Test whether the set is a true superset of other."""
        return other_set < self

    # SET Operations
    def union(self, key, *other_sets):
        """
        Return a new ``Set`` representing the union of *n* sets.

        :param key: String representing the key where to store the result (the union)
        :param other_sets: list of other ``Set``.
        :rtype: ``Set``

        >>> s1 = trol.Set('key1', redis)
        >>> s2 = trol.Set('key2', redis)
        >>> s1.add(['a', 'b', 'c'])
        3
        >>> s2.add(['d', 'e'])
        2
        >>> s3 = s1.union('key3', s2)
        >>> s3.key
        'key3'
        >>> s3.members == {'a', 'c', 'b', 'e', 'd'}
        True
        >>> s1.clear()
        >>> s2.clear()
        >>> s3.clear()

        """
        self.redis.sunionstore(key, [self.key] + [o.key for o in other_sets])
        return Set(key, redis=self.redis)

    def intersection(self, key, *other_sets):
        """
        Return a new ``Set`` representing the intersection of *n* sets.

        :param key: String representing the key where to store the result (the union)
        :param other_sets: list of other ``Set``.
        :rtype: trol.Set

        >>> s1 = trol.Set('key1', redis)
        >>> s2 = trol.Set('key2', redis)
        >>> s1.add(['a', 'b', 'c'])
        3
        >>> s2.add(['c', 'e'])
        2
        >>> s3 = s1.intersection('key3', s2)
        >>> s3.key
        'key3'
        >>> s3.members
        {'c'}
        >>> s1.clear()
        >>> s2.clear()
        >>> s3.clear()

        """

        self.redis.sinterstore(key, [self.key] + [o.key for o in other_sets])
        return Set(key, redis=self.redis)

    def difference(self, key, *other_sets):
        """
        Return a new ``Set`` representing the difference of *n* sets.

        :param key: String representing the key where to store the result (the union)
        :param other_sets: list of other ``Set``.
        :rtype: Set

        >>> s1 = trol.Set('key1', redis)
        >>> s2 = trol.Set('key2', redis)
        >>> s1.add(['a', 'b', 'c'])
        3
        >>> s2.add(['c', 'e'])
        2
        >>> s3 = s1.difference('key3', s2)
        >>> s3.key
        'key3'
        >>> s3.members == {'a', 'b'}
        True
        >>> s1.clear()
        >>> s2.clear()
        >>> s3.clear()

        """

        self.redis.sdiffstore(key, [self.key] + [o.key for o in other_sets])
        return Set(key, redis=self.redis)

    def update(self, *other_sets):
        """Update the set, adding elements from all other_sets.

        :param other_sets: list of ``Set``
        :rtype: None

        >>> s1 = trol.Set('key1', redis)
        >>> s1.add(['a', 'b', 'c'])
        3
        >>> s2 = trol.Set('key2', redis)
        >>> s2.add(['b', 'c', 'd'])
        3
        >>> s1.update(s2)
        >>> s1.members == {'a', 'b', 'c', 'd'}
        True
        >>> s1.clear()
        >>> s2.clear()

        """
        self.redis.sunionstore(
            self.key, [self.key] + [o.key for o in other_sets])

    def __ior__(self, other_set):
        self.redis.sunionstore(self.key, [self.key, other_set.key])
        return self

    def intersection_update(self, *other_sets):
        """
        Update the set, keeping only elements found in it and all other_sets.

        :param other_sets: list of ``Set``
        :rtype: None

        >>> s1 = trol.Set('key1', redis)
        >>> s1.add(['a', 'b', 'c'])
        3
        >>> s2 = trol.Set('key2', redis)
        >>> s2.add(['b', 'c', 'd'])
        3
        >>> s1.intersection_update(s2)
        >>> s1.members == {'b', 'c'}
        True
        >>> s1.clear()
        >>> s2.clear()

        """
        self.redis.sinterstore(
            self.key, [self.key] + [o.key for o in other_sets])

    def __iand__(self, other_set):
        self.redis.sinterstore(self.key, [self.key, other_set.key])
        return self

    def difference_update(self, *other_sets):
        """
        Update the set, removing elements found in others.

        :param other_sets: list of ``Set``
        :rtype: None

        >>> s1 = trol.Set('key1', redis)
        >>> s1.add(['a', 'b', 'c'])
        3
        >>> s2 = trol.Set('key2', redis)
        >>> s2.add(['b', 'c', 'd'])
        3
        >>> s1.difference_update(s2)
        >>> s1.members
        {'a'}
        >>> s1.clear()
        >>> s2.clear()

        """
        self.redis.sdiffstore(
            self.key, [self.key] + [o.key for o in other_sets])

    def __isub__(self, other_set):
        self.redis.sdiffstore(self.key, [self.key, other_set.key])
        return self

    def all(self):
        return {self.deserialize(m) for m in self.redis.smembers(self.key)}

    members = property(all)
    """
    Return the stored content of the Set.
    """

    def copy(self, key):
        """
        Copy the set to another key and return the new Set.

        .. WARNING::
            If the new key already contains a value, it will be overwritten.

        >>> s1 = trol.Set('key1', redis)
        >>> s1.add(['a', 'b', 'c'])
        3
        >>> s2 = s1.copy('key2')
        >>> s2.members == {"a", "b", "c"}
        True
        >>> s1.clear()
        >>> s2.clear()

        """

        copy = Set(key, redis=self.redis)
        self.redis.sunionstore(copy.key, self.key, self.key)
        return copy

    def __iter__(self):
        """Get an iterator for this set

        >>> s = trol.Set('test', redis)
        >>> s.add('a', 'b', 'b', 'c')
        3
        >>> for count, letter in enumerate(s):
        ...     print('{} Ah, Ah, Ah'.format(count+1))
        1 Ah, Ah, Ah
        2 Ah, Ah, Ah
        3 Ah, Ah, Ah
        >>> s.clear()

        """
        return self.members.__iter__()

    def sinter(self, *other_sets):
        """
        Performs an intersection between Sets and return the *RAW* result.

        .. NOTE::
            This function return an actual ``set`` object (from python) and not a ``Set``. See func:``intersection``.
        """
        return self.redis.sinter([self.key] + [self.key for s in other_sets])

    def sunion(self, *other_sets):
        """
        Performs a union between two sets and returns the *RAW* result.

        .. NOTE::
            This function return an actual ``set`` object (from python) and not a ``Set``.
        """
        return self.redis.sunion([self.key] + [self.key for s in other_sets])

    def sdiff(self, *other_sets):
        """
        Performs a difference between two sets and returns the *RAW* result.

        .. NOTE::
            This function return an actual ``set`` object (from python) and not a ``Set``.
            See function difference.

        """
        return self.redis.sdiff([self.key] + [self.key for s in other_sets])

    def scard(self):
        """
        Returns the cardinality of the Set.

        >>> s = trol.Set('test', redis)
        >>> s.add(['a', 'b', 'c'])
        3
        >>> s.scard()
        3
        >>> s.clear()

        :rtype: String containing the cardinality.

        """
        return self.redis.scard(self.key)

    def sismember(self, value):
        """
        Return ``True`` if the provided value is in the ``Set``.

        >>> s = trol.Set('test', redis)
        >>> s.add(['a', 'b', 'c'])
        3
        >>> s.sismember('d')
        False
        >>> s.clear()

        """
        return self.redis.sismember(self.key, self.serialize(value))

    def srandmember(self):
        """
        Return a random member of the set.

        >>> s = trol.Set('test', redis)
        >>> s.add(['a', 'b', 'c'])
        3
        >>> s.srandmember() in { 'a', 'b', 'c' }
        True
        >>> s.clear()

        """
        return self.deserialize(self.redis.srandmember(self.key))

    add = sadd
    """see sadd"""
    pop = spop
    """see spop"""
    remove = srem
    """see srem"""
    __contains__ = sismember
    __len__ = scard


class List(Collection):
    """
    This class represent a list object as seen in redis.
    """

    def all(self):
        """
        Returns all items in the list.
        """
        return self.lrange(0, -1)

    members = property(all)
    """Return all items in the list."""

    def llen(self):
        """
        Returns the length of the list.
        """
        return self.redis.llen(self.key)

    __len__ = llen

    def __getitem__(self, index):
        if isinstance(index, int):
            return self.lindex(index)
        elif isinstance(index, slice):
            indices = index.indices(len(self))
            return self.lrange(indices[0], indices[1] - 1)
        else:
            raise TypeError

    def __setitem__(self, index, value):
        self.lset(index, self.serialize(value))

    def lrange(self, start, stop):
        """
        Returns a range of items.

        :param start: integer representing the start index of the range
        :param stop: integer representing the size of the list.

        >>> l = trol.List('test', redis)
        >>> l.push(['a', 'b', 'c', 'd'])
        4
        >>> l.lrange(1, 2)
        ['b', 'c']
        >>> l.clear()

        """
        return [self.deserialize(v) for v in self.redis.lrange(self.key, start, stop)]

    def lpush(self, *values):
        """
        Push the value into the list from the *left* side

        :param values: a list of values or single value to push
        :rtype: long representing the number of values pushed.

        >>> l = trol.List('test', redis)
        >>> l.lpush(['a', 'b'])
        2
        >>> l.clear()

        """
        values = [self.serialize(v) for v in _parse_values(values)]
        return self.redis.lpush(self.key, *values)

    def rpush(self, *values):
        """
        Push the value into the list from the *right* side

        :param values: a list of values or single value to push
        :rtype: long representing the size of the list.

        >>> l = trol.List('test', redis)
        >>> l.lpush(['a', 'b'])
        2
        >>> l.rpush(['c', 'd'])
        4
        >>> l.members
        ['b', 'a', 'c', 'd']
        >>> l.clear()

        """

        values = [self.serialize(v) for v in _parse_values(values)]
        return self.redis.rpush(self.key, *values)

    def extend(self, iterable):
        """
        Extend list by appending elements from the iterable.

        :param iterable: an iterable objects.
        """
        self.rpush(*[self.serialize(e) for e in iterable])

    def count(self, value):
        """
        Return number of occurrences of value.

        :param value: a value tha *may* be contained in the list
        """
        return self.members.count(value)

    def lpop(self):
        """
        Pop the first object from the left.

        :return: the popped value.

        """
        return self.deserialize(self.redis.lpop(self.key))

    def rpop(self):
        """
        Pop the first object from the right.

        :return: the popped value.
        """
        return self.deserialize(self.redis.rpop(self.key))

    def rpoplpush(self, key):
        """
        Remove an element from the list,
        atomically add it to the head of the list indicated by key

        :param key: the key of the list receiving the popped value.
        :return: the popped (and pushed) value

        >>> l = trol.List('list1', redis)
        >>> l.push(['a', 'b', 'c'])
        3
        >>> l.rpoplpush('list2')
        'c'
        >>> l2 = trol.List('list2', redis)
        >>> l2.members
        ['c']
        >>> l.clear()
        >>> l2.clear()

        """
        return self.deserialize(self.redis.rpoplpush(self.key, key))

    def lrem(self, value, num=1):
        """
        Remove first occurrence of value.

        :return: 1 if the value has been removed, 0 otherwise
        """
        return self.redis.lrem(self.key, value, num)

    def reverse(self):
        """
        Reverse the list in place.

        ..NOTE:
            This command must make two network calls, transfering the list each way
            It is still atomic though

        :return: None
        """
        def reversefn(pipe):
            values = pipe.lrange(self.key, 0, -1)
            pipe.multi()
            pipe.delete(self.key)
            pipe.lpush(self.key, values)

        self.redis.transaction(reversefn, self.key)

    def copy(self, key):
        """Copy the list to a new list.

        ..WARNING:
            If destination key already contains a value, it clears it before copying.
        """
        def copyfn(pipe):
            values = pipe.lrange(self.key, 0, -1)
            pipe.multi()
            pipe.delete(key)
            pipe.rpush(key, values)

        self.redis.transaction(copyfn, self.key, key)
        copy = List(key, self.redis)
        return copy

    def ltrim(self, start, end):
        """
        Trim the list from start to end.

        :return: None
        """
        return self.redis.ltrim(self.key, start, end)

    def lindex(self, idx):
        """
        Return the value at the index *idx*

        :param idx: the index to fetch the value.
        :return: the value or None if out of range.

        """
        return self.deserialize(self.redis.lindex(self.key, idx))

    def lset(self, idx, value):
        """
        Set the value in the list at index *idx*

        :return: True is the operation succeed.

        >>> l = trol.List('test', redis)
        >>> l.push(['a', 'b', 'c'])
        3
        >>> l.lset(0, 'e')
        True
        >>> l.members
        ['e', 'b', 'c']
        >>> l.clear()

        """
        return self.redis.lset(self.key, idx, self.serialize(value))

    def __iter__(self):
        """Get an iterator for this list

        >>> l = trol.List('test', redis)
        >>> l.lpush('a', 'b', 'c')
        3
        >>> for letter in l:
        ...     print(letter)
        c
        b
        a
        >>> l.clear()

        """
        return self.members.__iter__()

    def __repr__(self):
        """Get the string representation of this set

        >>> l = trol.List('test', redis)
        >>> repr(l)
        "<List 'test'>"

        """
        return "<%s '%s'>" % (self.__class__.__name__, self.key)

    __len__ = llen
    remove = lrem
    trim = ltrim
    shift = lpop
    unshift = lpush
    pop = rpop
    pop_onto = rpoplpush
    push = rpush
    append = rpush


class SortedSet(Collection):
    """
    This class represents a SortedSet in redis.
    Use it if you want to arrange your set in any order.

    """

    def __getitem__(self, index):
        if isinstance(index, slice):
            return [self.deserialize(v) for v in self.zrange(index.start, index.stop)]
        else:
            return self.deserialize(self.zrange(index, index)[0])

    def score(self, member):
        """
        Returns the score of member.
        """
        return self.zscore(member)

    def __contains__(self, val):
        return self.zscore(val) is not None

    @property
    def members(self):
        """
        Returns the members of the set.
        """
        return self.zrange(0, -1)

    @property
    def revmembers(self):
        """
        Returns the members of the set in reverse.
        """
        return [self.deserialize(v) for v in self.zrevrange(0, -1)]

    def __iter__(self):
        return self.members.__iter__()

    def __reversed__(self):
        return self.revmembers.__iter__()

    @property
    def _min_score(self):
        """
        Returns the minimum score in the SortedSet.
        """
        try:
            return self.zscore(self.__getitem__(0))
        except IndexError:
            return None

    @property
    def _max_score(self):
        """
        Returns the maximum score in the SortedSet.
        """
        try:
            return self.zscore(self.__getitem__(-1))
        except IndexError:
            return None

    def lt(self, v, limit=None, offset=None):
        """
        Returns the list of the members of the set that have scores
        less than v.

        :param v: the score to compare to.
        :param limit: limit the result to *n* elements
        :param offset: Skip the first *n* elements
        """
        if limit is not None and offset is None:
            offset = 0
        return self.zrangebyscore("-inf", "(%f" % v,
                                  start=offset, num=limit)

    def le(self, v, limit=None, offset=None):
        """
        Returns the list of the members of the set that have scores
        less than or equal to v.

        :param v: the score to compare to.
        :param limit: limit the result to *n* elements
        :param offset: Skip the first *n* elements

        """
        if limit is not None and offset is None:
            offset = 0
        return self.zrangebyscore("-inf", v,
                                  start=offset, num=limit)

    def gt(self, v, limit=None, offset=None, withscores=False):
        """Returns the list of the members of the set that have scores
        greater than v.
        """
        if limit is not None and offset is None:
            offset = 0
        return self.zrangebyscore("(%f" % v, "+inf",
                                  start=offset, num=limit, withscores=withscores)

    def ge(self, v, limit=None, offset=None, withscores=False):
        """Returns the list of the members of the set that have scores
        greater than or equal to v.

        :param v: the score to compare to.
        :param limit: limit the result to *n* elements
        :param offset: Skip the first *n* elements

        """
        if limit is not None and offset is None:
            offset = 0
        return self.zrangebyscore("%f" % v, "+inf",
                                  start=offset, num=limit, withscores=withscores)

    def between(self, min, max, limit=None, offset=None):
        """
        Returns the list of the members of the set that have scores
        between min and max.

        .. Note::
            The min and max are inclusive when comparing the values.

        :param min: the minimum score to compare to.
        :param max: the maximum score to compare to.
        :param limit: limit the result to *n* elements
        :param offset: Skip the first *n* elements

        >>> s = trol.SortedSet('foo', redis)
        >>> s.add('a', 10)
        1
        >>> s.add('b', 20)
        1
        >>> s.add('c', 30)
        1
        >>> s.between(20, 30)
        ['b', 'c']
        >>> s.clear()

        """
        if limit is not None and offset is None:
            offset = 0
        return self.zrangebyscore(min, max,
                                  start=offset, num=limit)

    def zadd(self, members, score=1):
        """
        Add members in the set and assign them the score.

        :param members: a list of item or a single item
        :param score: the score the assign to the item(s)

        >>> s = trol.SortedSet('foo', redis)
        >>> s.add('a')
        1
        >>> s.zscore('a')
        1.0
        >>> s.add('b', 20)
        1
        >>> s.zscore('b')
        20.0
        >>> s.add({'c':5, 'd':6})
        2
        >>> s.clear()

        """
        _members = []
        if not isinstance(members, dict):
            _members = [self.serialize(members), score]
        else:
            for member, score in members.items():
                _members += [self.serialize(member), score]

        return self.redis.zadd(self.key, *_members)

    def zrem(self, *values):
        """
        Remove the values from the SortedSet

        :return: True if **at least one** value is successfully
                 removed, False otherwise

        >>> s = trol.SortedSet('foo', redis)
        >>> s.add('a', 10)
        1
        >>> s.zrem('a')
        1
        >>> s.members
        []
        >>> s.clear()

        """
        values = [self.serialize(v) for v in _parse_values(values)]
        return self.redis.zrem(self.key, *values)

    def zincrby(self, att, value=1):
        """
        Increment the score of the item by ``value``

        :param att: the member to increment
        :param value: the value to add to the current score
        :returns: the new score of the member

        >>> s = trol.SortedSet('foo', redis)
        >>> s.add('a', 10)
        1
        >>> s.zincrby('a', 10)
        20.0
        >>> s.clear()

        """
        return self.redis.zincrby(self.key, self.serialize(att), value)

    def zrevrank(self, member):
        """
        Returns the ranking in reverse order for the member

        >>> s = trol.SortedSet('foo', redis)
        >>> s.add('a', 10)
        1
        >>> s.add('b', 20)
        1
        >>> s.revrank('a')
        1
        >>> s.clear()

        """
        return self.redis.zrevrank(self.key, self.serialize(member))

    def zrange(self, start, stop, withscores=False):
        """
        Returns all the elements including between ``start`` (non included) and
        ``stop`` (included).

        :param withscore: True if the score of the elements should
                          also be returned

        >>> s = trol.SortedSet('foo', redis)
        >>> s.add('a', 10)
        1
        >>> s.add('b', 20)
        1
        >>> s.add('c', 30)
        1
        >>> s.zrange(1, 3)
        ['b', 'c']
        >>> s.zrange(1, 3, withscores=True)
        [('b', 20.0), ('c', 30.0)]
        >>> s.clear()

        """
        if withscores:
            return [(self.deserialize(v), s) for (v, s) in self.redis.zrange(self.key, start, stop, withscores=True)]
        else:
            return [self.deserialize(v) for v in self.redis.zrange(self.key, start, stop, withscores=False)]

    def zrevrange(self, start, end, **kwargs):
        """
        Returns the range of items included between ``start`` and ``stop``
        in reverse order (from high to low)

        >>> s = trol.SortedSet('foo', redis)
        >>> s.add('a', 10)
        1
        >>> s.add('b', 20)
        1
        >>> s.add('c', 30)
        1
        >>> s.zrevrange(1, 2)
        ['b', 'a']
        >>> s.clear()

        """
        return[self.deserialize(v) for v in self.redis.zrevrange(self.key, start, end, **kwargs)]

    def zrangebyscore(self, min, max, **kwargs):
        """
        Returns the range of elements included between the scores (min and max)

        >>> s = trol.SortedSet('foo', redis)
        >>> s.add('a', 10)
        1
        >>> s.add('b', 20)
        1
        >>> s.add('c', 30)
        1
        >>> s.zrangebyscore(20, 30)
        ['b', 'c']
        >>> s.clear()

        """
        return [self.deserialize(v) for v in self.redis.zrangebyscore(self.key, min, max, **kwargs)]

    def zrevrangebyscore(self, max, min, **kwargs):
        """
        Returns the range of elements included between the scores (min and max)

        >>> s = trol.SortedSet('foo', redis)
        >>> s.add('a', 10)
        1
        >>> s.add('b', 20)
        1
        >>> s.add('c', 30)
        1
        >>> s.zrangebyscore(20, 20)
        ['b']
        >>> s.clear()

        """
        return [self.deserialize(v) for v in self.redis.zrevrangebyscore(self.key, max, min, **kwargs)]

    def zcard(self):
        """
        Returns the cardinality of the SortedSet.

        >>> s = trol.SortedSet('foo', redis)
        >>> s.add("a", 1)
        1
        >>> s.add("b", 2)
        1
        >>> s.add("c", 3)
        1
        >>> s.zcard()
        3
        >>> s.clear()

        """
        return self.redis.zcard(self.key)

    def zscore(self, elem):
        """
        Return the score of an element

        >>> s = trol.SortedSet('foo', redis)
        >>> s.add("a", 10)
        1
        >>> s.score("a")
        10.0
        >>> s.clear()

        """
        return self.redis.zscore(self.key, self.serialize(elem))

    def zremrangebyrank(self, start, stop):
        """
        Remove a range of element between the rank ``start`` and
        ``stop`` both included.

        :return: the number of item deleted

        >>> s = trol.SortedSet('foo', redis)
        >>> s.add("a", 10)
        1
        >>> s.add("b", 20)
        1
        >>> s.add("c", 30)
        1
        >>> s.zremrangebyrank(1, 2)
        2
        >>> s.members
        ['a']
        >>> s.clear()

        """
        return self.redis.zremrangebyrank(self.key, start, stop)

    def zremrangebyscore(self, min_value, max_value):
        """
        Remove a range of element by between score ``min_value`` and
        ``max_value`` both included.

        :returns: the number of items deleted.

        >>> s = trol.SortedSet('foo', redis)
        >>> s.add("a", 10)
        1
        >>> s.add("b", 20)
        1
        >>> s.add("c", 30)
        1
        >>> s.zremrangebyscore(10, 20)
        2
        >>> s.members
        ['c']
        >>> s.clear()

        """

        return self.redis.zremrangebyscore(self.key, min_value, max_value)

    def zrank(self, elem):
        """
        Returns the rank of the element.

        >>> s = trol.SortedSet('foo', redis)
        >>> s.add({"a": 30, 'b':20, 'c':10})
        3
        >>> s.zrank("a")
        2
        >>> s.clear()

        """
        return self.redis.zrank(self.key, self.serialize(elem))

    def eq(self, value):
        """
        Returns the elements that have ``value`` for score.
        """
        return self.zrangebyscore(value, value)

    __len__ = zcard
    revrank = zrevrank
    score = zscore
    rank = zrank
    incr_by = zincrby
    add = zadd
    remove = zrem


class Hash(Collection, collections.MutableMapping):

    def __iter__(self):
        return self.hgetall().__iter__()

    def __repr__(self):
        """Gets the string representation of this object

        >>> h = trol.Hash('test', redis)
        >>> repr(h)
        "<Hash 'test'>"

        """
        return "<%s '%s'>" % (self.__class__.__name__, self.key)

    def _set_dict(self, new_dict):
        self.clear()
        self.update(new_dict)

    def hlen(self):
        """Returns the number of elements in the Hash.
        """
        return self.redis.hlen(self.key)

    def hset(self, member, value):
        """
        Set ``member`` in the Hash at ``value``.

        :returns: 1 if member is a new field and the value has been
                  stored, 0 if the field existed and the value has been
                  updated.

        >>> h = trol.Hash('foo', redis)
        >>> h.hset("bar", "value")
        1
        >>> h.clear()

        """
        return self.redis.hset(self.key, member, self.serialize(value))

    def hdel(self, *hkeys):
        """
        Delete one or more hash fields by key.

        :param hkeys: on or more fields to remove.
        :return: the number of fields that were removed

        >>> h = trol.Hash('foo', redis)
        >>> h.hset("bar", "value")
        1
        >>> h.hdel("bar")
        1
        >>> h.clear()

        """
        return self.redis.hdel(self.key, *hkeys)

    def hkeys(self):
        """
        Returns all fields name in the Hash
        """
        return self.redis.hkeys(self.key)

    def hgetall(self):
        """
        Returns all the fields and values in the Hash.

        :rtype: dict
        """
        return dict([(k, self.deservialize(v)) for (k, v) in self.redis.hgetall(self.key)])

    def hvals(self):
        """
        Returns all the values in the Hash

        :rtype: list
        """
        return [self.deserialize(v) for v in self.redis.hvals(self.key)]

    def hget(self, field):
        """
        Returns the value stored in the field, None if the field doesn't exist.
        """
        return self.deserialize(self.redis.hget(self.key, field))

    def hexists(self, field):
        """
        Returns ``True`` if the field exists, ``False`` otherwise.
        """
        return self.redis.hexists(self.key, field)

    def hincrby(self, field, increment=1):
        """
        Increment the value of the field.
        :returns: the value of the field after incrementation

        >>> h = trol.Hash('foo', redis)
        >>> h.hincrby("bar", 10)
        10
        >>> h.hincrby("bar", 2)
        12
        >>> h.clear()

        """
        return self.redis.hincrby(self.key, field, increment)

    def hmget(self, fields):
        """
        Returns the values stored in the fields.
        """
        return [self.deserialize(v) for v in self.redis.hmget(self.key, fields)]

    def hmset(self, mapping):
        """
        Sets or updates the fields with their corresponding values.

        :param mapping: a dict with keys and values
        """
        mapping = dict([(k, self.serialize(v)) for (k, v) in mapping])
        return self.redis.hmset(self.key, mapping)

    keys = hkeys
    values = hvals
    _get_dict = hgetall
    __getitem__ = hget
    __setitem__ = hset
    __delitem__ = hdel
    __len__ = hlen
    __contains__ = hexists
    dict = property(_get_dict, _set_dict)



from trol import ModelType, Collection, Property 

class DatabaseType(type):
    def __init__(cls, *args, **kwargs):
        """
        TODO: Write a docstring
        """
        cls._rtol_properties = dict()
        cls._rtol_collections = dict()
        cls._rtol_models = dict()

        for attrname, attr in cls.__dict__.items():
            if isinstance(attr, Property):
                cls._rtol_properties[attrname] = attr

                if attr._name is None:
                    attr._name = attrname

            if isinstance(attr, Collection):
                cls._rtol_collections[attrname] = attr

                if attr._name is None:
                    attr._name = attrname

                if attr._redis is None:
                    attr._redis = getattr(cls, 'redis', None)

            if isinstance(attr, ModelType):
                cls._rtol_models[attrname] = attr
                attr._rtol_database = cls

        super().__init__(*args, **kwargs)

class Database(metaclass=DatabaseType):
    # Wait a minute! This class is empty
    # Ya, but it gives you the nice syntactic sugar of:
    #     class MyDB(Database):
    #
    # and avoids the question "What's a metaclass?"
    redis = None



import unittest
from trol import highlander


@highlander()
class Z:
    def __init__(self):
        pass


@highlander('alpha')
class A:
    def __init__(self, alpha):
        pass


@highlander('alpha', 'beta')
class B:
    def __init__(self, alpha, beta):
        pass


@highlander('alpha', 'beta')
class C:
    def __init__(self, alpha, beta, gamma=None):
        pass


class HighlanderTests(unittest.TestCase):

    def test_obtain_returns_instance(self):
        self.assertIsInstance(A.obtain("X"), A)
        self.assertIsInstance(B.obtain("X", "Y"), B)
        self.assertIsInstance(C.obtain("X", "Y"), C)
        self.assertIsInstance(Z.obtain(), Z)

    def test_obtain_returns_shared_instance(self):
        z1, z2 = Z.obtain(), Z.obtain()
        self.assertIs(z1, z2)

        a1, a2 = A.obtain(1), A.obtain(1)
        self.assertIs(a1, a2)

        b1, b2 = B.obtain(1, "X"), B.obtain(1, "X")
        self.assertIs(b1, b2)

        c1, c2 = C.obtain(1, "X"), C.obtain(1, "X")
        self.assertIs(c1, c2)

    def test_obtain_returns_unique_instance(self):
        a1, a2 = A.obtain(1), A.obtain(2)
        self.assertIsNot(a1, a2)

        b1, b2 = B.obtain(1, "X"), B.obtain(2, "X")
        self.assertIsNot(b1, b2)

        c1, c2 = C.obtain(1, "X"), C.obtain(1, "Y")
        self.assertIsNot(c1, c2)

    def test_obtain_raises_type_error(self):
        with self.assertRaises(TypeError):
            A.obtain("too", "many", "arguments")

        with self.assertRaises(TypeError):
            B.obtain("too_many_arguments")

    def test_obtain_calls_init(self):
        @highlander('alpha', 'beta')
        class X:
            def __init__(this, alpha, beta):
                this.alpha = alpha
                this.beta = beta

        keyA, keyB = "keyA", 5
        x1 = X.obtain(keyA, keyB)

        self.assertIs(x1.alpha, keyA)
        self.assertIs(x1.beta, keyB)

    def test_obtain_calls_init_once(self):
        @highlander()
        class X:
            def __init__(this):
                self.assertFalse(hasattr(this, 'mark'))
                this.mark = None

        x1 = X.obtain()
        x2 = X.obtain()



import doctest
import unittest
import trol
from redis import Redis
from .common import ensure_redis_is_online

container_token = None
redis = None


def setUp(dtest):
    global container_token
    global redis

    container_token = ensure_redis_is_online()
    redis = Redis('localhost')


def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocFileSuite(
        'model.py', 'collection.py', 'util.py', 'README.rst', package=trol, setUp=setUp))
    return tests



from weakref import WeakValueDictionary


def highlander(*attrs):
    """A decorator which enables class level object caching

    Add a class level cache and an obtain function which allows retrieval
    This is used for objects where only one needs to or should exist at a time and that object can be shared safely among threads

    "There can only be one!"

    Args:
        *attrs (list[str]): The attribute names which uniquely identify the object in the cache.
             These are also the arguments which will be passed to __init__ when building the cached object in `obtain`

    Returns:
        Callable[[type], type]: The function which will be applied to decorate the class
    """
    def decoratorfn(cls):
        cls._highlander_cache = WeakValueDictionary()

        @classmethod
        def obtain(cls, *args, **kwargs):
            arglen = len(args) + len(kwargs)
            if arglen != len(attrs):
                raise TypeError("obtain() takes exactly {} argument(s) but {} were given".format(
                    len(attrs), arglen))

            # Build the identifier by adding positional args to a list, then using attrs to disern an order for kwargs
            identifier = [*args]
            for attr in attrs[len(args):]:
                identifier.append(kwargs[attr])
            identifier = tuple(identifier)  # FREEZE!

            try:
                return cls._highlander_cache[identifier]
            except KeyError:
                inst = cls(*args, **kwargs)
                cls._highlander_cache[identifier] = inst
                return inst

        obtain.__func__.__doc__ = \
            """Retrieves or builds an instance of {cls.__name__} uniquely identified by {attrs}

        If the object must be built, the arguments to this function will be passed to __init__

        Args:
            {attrsfmt}

        Returns:
            {cls.__name__}: Obtained from the cache or built and stored there
        """.format(attrs=attrs, cls=cls, attrsfmt='\n            '.join(attrs))

        cls.obtain = obtain
        return cls

    return decoratorfn



import unittest
import ddt
import docker
from .common import ensure_redis_is_online
from redis import StrictRedis
from trol import Serializer, Deserializer, Model


@ddt.ddt
class OnlineUtilTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.container_token = ensure_redis_is_online()
        cls.redis = StrictRedis(host='localhost', port=6379, db=0)

    def setUp(self):
        self.redis.flushall()

    @ddt.data(
        (str, "I am a test string!"),
        (int, 42),
        (float, 3.1415),
        (bytes, b'\xDE\xAD\xBE\xEF')
    )
    @ddt.unpack
    def test_roundtrip(self, typ, data):
        key = "key"

        s = Serializer(typ)
        d = Deserializer(typ)

        self.redis.set(key, s(data))
        result = d(self.redis.get(key))

        self.assertEqual(data, result)


class A(Model):
    pass


class B(Model):
    def __init__(self):
        self.model_name = 'BAR'
        self.id = 'x'


class C(Model):
    def __init__(self):
        self.key = 'foobar'


@ddt.ddt
class OfflineUtilTests(unittest.TestCase):

    def test_serialize_model(self):
        s = Serializer(Model)
        d = Deserializer(Model)
        a = A()

        replicon = d(s(a))
        self.assertIsInstance(a, A)
        self.assertEquals(a.model_name, replicon.model_name)

        with self.assertRaises(AttributeError):
            a.id

        with self.assertRaises(AttributeError):
            a.key

        b = B()

        replicon = d(s(b))

        self.assertIsInstance(b, B)
        self.assertEquals(b.model_name, replicon.model_name)
        self.assertEquals(b.id, replicon.id)
        self.assertEquals(b.key, replicon.key)

        c = C()

        replicon = d(s(c))

        self.assertIsInstance(c, C)
        self.assertEquals(c.model_name, replicon.model_name)
        with self.assertRaises(AttributeError):
            a.id
        self.assertEquals(c.key, replicon.key)



import unittest
import doctest
from redis import StrictRedis
from .common import ensure_redis_is_online
from trol import Model, ModelType, Property


class Alpha(Model):
    pass


class OfflineModelTests(unittest.TestCase):

    def test_key_retrieval(self):
        a = Alpha()

        with self.assertRaises(AttributeError):
            a.key

        a.id = "a"
        self.assertEquals(a.key, "Alpha:a")

        a.model_name = "FOO"
        self.assertEquals(a.key, "FOO:a")

    def test_redis_retrieval(self):
        a = Alpha()

        self.assertIsNone(a.redis)

        canary = object()
        a.redis = canary
        self.assertIs(a.redis, canary)


class X(Model):
    def __init__(self, id):
        self.id = id

    one = Property(autocommit=False, typ=bytes)
    two = Property('2', typ=str)
    three = Property(alwaysfetch=True, typ=int)
    four = Property(autocommit=False, typ=float)


class OnlineModelTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.container_token = ensure_redis_is_online()
        X.redis = StrictRedis(host='localhost', port=6379, db=0)

    def setUp(self):
        X.redis.flushall()

    def test_invalidate(self):
        x = X("xyz")

        X.redis.set("X:xyz:one", b'\xDE\xAD\xBE\xEF')
        X.redis.set("X:xyz:2", "canary")
        X.redis.set("X:xyz:three", 42)
        X.redis.set("X:xyz:four", 3.14)

        self.assertEquals(x.one, b'\xDE\xAD\xBE\xEF')
        self.assertEquals(x.two,  "canary")
        self.assertEquals(x.three, 42)
        self.assertEquals(x.four, 3.14)

        X.redis.set("X:xyz:one", b'\x8B\xAD\xF0\x0D')
        X.redis.set("X:xyz:2", "swallow")
        X.redis.set("X:xyz:three", 1024)
        X.redis.set("X:xyz:four", 2.71828)

        self.assertEquals(x.one, b'\xDE\xAD\xBE\xEF')
        self.assertEquals(x.two,  "canary")
        self.assertEquals(x.three, 1024)
        self.assertEquals(x.four, 3.14)

        x.invalidate('two', 'one')

        self.assertEquals(x.one, b'\x8B\xAD\xF0\x0D')
        self.assertEquals(x.two, "swallow")
        self.assertEquals(x.three, 1024)
        self.assertEquals(x.four, 3.14)

        x.invalidate()

        self.assertEquals(x.one, b'\x8B\xAD\xF0\x0D')
        self.assertEquals(x.two, "swallow")
        self.assertEquals(x.three, 1024)
        self.assertEquals(x.four, 2.71828)

    def test_commit(self):
        x = X("xyz")

        x.one = b'\xDE\xAD\xBE\xEF'
        x.two = "canary"
        x.three = 42
        x.four = 3.14

        self.assertIsNone(X.redis.get("X:xyz:one"))
        self.assertEquals(X.redis.get("X:xyz:2").decode('utf-8'),  "canary")
        self.assertEquals(X.redis.get("X:xyz:three").decode('utf-8'), '42')
        self.assertIsNone(X.redis.get("X:xyz:four"))

        x.commit('one')

        self.assertEquals(X.redis.get("X:xyz:one"), b'\xDE\xAD\xBE\xEF')
        self.assertEquals(X.redis.get("X:xyz:2").decode('utf-8'),  "canary")
        self.assertEquals(X.redis.get("X:xyz:three").decode('utf-8'), '42')
        self.assertIsNone(X.redis.get("X:xyz:four"))

        x.commit()

        self.assertEquals(X.redis.get("X:xyz:one"), b'\xDE\xAD\xBE\xEF')
        self.assertEquals(X.redis.get("X:xyz:2").decode('utf-8'),  "canary")
        self.assertEquals(X.redis.get("X:xyz:three").decode('utf-8'), '42')
        self.assertEquals(X.redis.get("X:xyz:four").decode('utf-8'), '3.14')

    def test_delete(self):
        x = X("xyz")

        x.one = b'\xDE\xAD\xBE\xEF'
        x.two = "canary"
        x.three = 42
        x.four = 3.14
        x.commit()

        self.assertEquals(X.redis.get("X:xyz:one"), b'\xDE\xAD\xBE\xEF')
        self.assertEquals(X.redis.get("X:xyz:2").decode('utf-8'),  "canary")
        self.assertEquals(X.redis.get("X:xyz:three").decode('utf-8'), '42')
        self.assertEquals(X.redis.get("X:xyz:four").decode('utf-8'), '3.14')

        x.delete('two', 'three')

        self.assertEquals(X.redis.get("X:xyz:one"), b'\xDE\xAD\xBE\xEF')
        self.assertIsNone(X.redis.get("X:xyz:2"))
        self.assertIsNone(X.redis.get("X:xyz:three"))
        self.assertEquals(X.redis.get("X:xyz:four").decode('utf-8'), '3.14')

        x.delete()

        self.assertIsNone(X.redis.get("X:xyz:one"))
        self.assertIsNone(X.redis.get("X:xyz:2"))
        self.assertIsNone(X.redis.get("X:xyz:three"))
        self.assertIsNone(X.redis.get("X:xyz:four"))

    def test_update(self):
        x = X("xyz")

        x.update(
            one=b'\xDE\xAD\xBE\xEF',
            two="canary",
            three=42
        )

        self.assertEquals(X.one.value(x), b'\xDE\xAD\xBE\xEF')
        self.assertEquals(X.two.value(x),  "canary")
        self.assertEquals(X.three.value(x), 42)
        self.assertIs(X.four.value(x), X.four.null)

        self.assertIsNone(X.redis.get("X:xyz:one"))
        self.assertEquals(X.redis.get("X:xyz:2").decode('utf-8'),  "canary")
        self.assertEquals(X.redis.get("X:xyz:three").decode('utf-8'), '42')
        self.assertIsNone(X.redis.get("X:xyz:four"))



"""A set of utility functions for use in multiple parts of trol

This module currently contains the serialize and deserialize methods for the Property and Collection classes I implement
"""

import weakref
import trol

# redis-py did this for me already, no need to reinvent the wheel :P


def serialize_str(obj):
    return obj


def serialize_int(obj):
    return obj


def serialize_float(obj):
    return obj


def serialize_bytes(obj):
    return obj


serializers = {
    str: serialize_str,
    int: serialize_int,
    float: serialize_float,
    bytes: serialize_bytes,
}
""" dict[type, Callable[[object], bytes]]: A dictionary of serializers known trol classes

Additonal entries can be added to support new serializable types

>>> import redis
>>> import trol
>>> class HotNewClass:
...     def __init__(self, howhot):
...         self.howhot = howhot
...
>>> @trol.serializer(HotNewClass)
... def hotnew_serializer(hnc):
...     print("HNC IS BEING SERIALIZED!")
...     return '<HOT>{}'.format(hnc.howhot)
...
>>> @trol.deserializer(HotNewClass)
... def hotnew_deserializer(byts):
...     print("RETURN OF THE HNC!")
...     howhot = int(byts.decode('utf-8').strip('<HOT>'))
...     return HotNewClass(howhot)
...
>>> class SweetModel(trol.Model):
...     def __init__(self, ident, redis):
...         self.id = ident
...         self.redis = redis
...
...     bar = trol.Property(typ=HotNewClass)
...
>>> r = redis.Redis('localhost')
>>> sm = SweetModel('xyz', r)
>>> sm.bar = HotNewClass(10)
HNC IS BEING SERIALIZED!
>>> r.get(sm.key + ':bar')
b'<HOT>10'
>>> sm.invalidate()
>>> sm.bar.howhot
RETURN OF THE HNC!
10

"""

def serializer(cls):
    """A convinience decorator to register a serializer"""
    def decorator(f):
        serializers[cls] = f
        return f
    return decorator

class Serializer:
    """A class containing the provided serialize functions for selected type

    The serializers here are aim for human-readability, which means they are not their most optimized
    If you want better performance through avoidance to and from `str`, implement a new serializer

    Attributes:
        func (Callable[[object], object]: The serialization function this class is constructed to use
            The input and return types are determined by the type used to construct this object
            The output type will be a serial type accepted by redis-py

    Args:
        typ (type): The type of objects which will be serialized by this Serializer
            Currently supported types are: `str`, `int`, `float`, and `bytes`
    """

    def __init__(self, typ):
        try:
            self.func = serializers[typ]
        except KeyError:
            raise ValueError("{} is not supported".format(typ.__name__))

    def __call__(self, obj):
        """Serialize the input object into a form acceptable to redis-py

        Args:
            obj (object): An object of the same type used to construct this Serializer

        Returns:
            object: A serialized version acceptable to redis-py
        """
        return self.func(obj)


def deserialize_str(byts):
    return byts.decode('utf-8')


def deserialize_int(byts):
    return int(byts.decode('utf-8'))


def deserialize_float(byts):
    return float(byts.decode('utf-8'))


def deserialize_bytes(byts):
    return byts

deserializers = {
    str: deserialize_str,
    int: deserialize_int,
    float: deserialize_float,
    bytes: deserialize_bytes,
}
""" dict[type, Callable[[bytes], object]]: A dictionary of deserializers known trol classes

Additonal entries can be added to support new deserializable types
There should be an entry here for each one in serializers
"""

def deserializer(cls):
    """A convinience decorator to register a deserializer"""
    def decorator(f):
        deserializers[cls] = f
        return f
    return decorator


class Deserializer:
    """A class containing the provided deserialize functions for selected type

    The deserializers here are aim for human-readability, which means they are not their most optimized
    If you want better performance through avoidance to and from `str`, implement a new deserializer

    Attributes:
        func (Callable[[bytes], object]: The deserialization function this class is constructed to use
            The return type is determined by the type used to construct this object

    Args:
        typ (type): The type of objects which will be deserialized by this Deserializer
            Currently supported types are: `str`, `int`, `float`, and `bytes`
    """

    def __init__(self, typ):
        try:
            self.func = deserializers[typ]
        except KeyError:
            raise ValueError("{} is not supported".format(typ.__name__))

    def __call__(self, obj):
        """Deserialize the input bytes from redis-py into the desired object type

        Args:
            byts (bytes): A bytes object containing the data you want to deserialize

        Returns:
            object: An object of the desired type
        """
        return self.func(obj)



from distutils.core import setup

setup(
  name = 'trol',
  packages = ['trol'], # this must be the same as the name above
  version = '0.1.1',
  description = 'A light and predictable Redis object mapper',
  author = 'Victor "Nate" Graf',
  author_email = 'nategraf1@gmail.com',
  license = 'MIT',
  url = 'https://github.com/nategraf/trol', # use the URL to the github repo
  download_url = 'https://github.com/nategraf/trol/archive/0.1.1.tar.gz',
  keywords = ['redis', 'object', 'mapper', 'thin', 'rom', 'redisco'], # arbitrary keywords
  install_requires=['redis'],
  python_requires='>=3',
  classifiers=[
    'Development Status :: 3 - Alpha',

    # Indicate who your project is intended for
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',

    # Pick your license as you wish (should match "license" above)
    'License :: OSI Approved :: MIT License',

    'Operating System :: OS Independent',
    'Natural Language :: English',

    # Specify the Python versions you support here. In particular, ensure
    # that you indicate whether you support Python 2, Python 3 or both.
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3 :: Only',

    'Topic :: Database',
    'Topic :: Software Development :: Libraries :: Python Modules'
  ]
)



class RedisKeyError(Exception):
    """An exception to indicate that a Redis key does not exist (i.e. redis.get returned None)

    Attributes:
        key (str): The Redis key which was triggered the exception

    Args:
        key (str): Sets the `key` attribute
    """

    def __init__(self, key):
        self.key = key

    def __str__(self):
        return "Key '{}' does not exist in Redis".format(self.key)
